﻿Imports System.IO

Public Class NodeFillJob
    ' Die Objekt-Methoden GetSubFolders() und Synchronize() trennen die 
    '  Befüllung eines Treenodes mit Directories in 2 Schritte auf: 
    '   GetSubFolders ermittelt nur die DirectoryNames. Das ist zeitaufwändig, 
    '    aber kann im NebenThread geschehen. 
    '   Synchronize() füllt dann den Node mit Subnodes diesen Names. Das muss
    '    im Gui-Thread erfolgen, geht dafür recht flott.
    ' Die Public Shared Methoden befüllen den Node direkt, ohne ein 
    '  NodeFillJob-Objekt zu erstellen und ohne Threading.

    ' Es gibt gewisse Ähnlichkeiten zum command-pattern
    ' (s.a.: "http://de.wikipedia.org/wiki/Kommando_%28Entwurfsmuster%29")

    Private ReadOnly _Node As TreeNode
    Private _SubFolders As List(Of String)

    Public Sub New(ByVal Node As TreeNode)
        _Node = Node
    End Sub

    Public Sub GetSubFolders()
        ' Für Aufruf aus den Nebenthread, da GetSubFolders zeitaufwändig ist.
        _SubFolders = GetSubFolders(_Node.FullPath)
    End Sub

    Private Shared Function GetSubFolders( _
            ByVal ParentPath As String) As List(Of String)

        Try
            Dim Dirs = Directory.GetDirectories(ParentPath & _
                                                 Path.DirectorySeparatorChar)
            Return Dirs.Select(Function(Dir As String) _
                                   Path.GetFileName(Dir)).ToList
        Catch ex As System.UnauthorizedAccessException
            ' Bei bestimmten Directories wirft das Betriebssystem diese 
            '  Exception, wenn ihr Inhalt eingesehen werden soll.
            Return New List(Of String) ' gib leere Liste zurück
        End Try
    End Function

    Public Sub Synchronize()
        ' Für Aufruf nach Rückkehr aus dem Nebenthread, da auf die TreeNodes 
        '  nicht im NebenThread zugegriffen werden kann.
        Synchronize(_Node.Nodes, _SubFolders)
        _Node.Tag = Nothing ' als synchronisiert markieren
    End Sub

    Private Shared Sub Synchronize(ByVal Nodes As TreeNodeCollection, _
                                     ByVal Names As List(Of String))
        ' Sind schon Nodes vorhanden, gibt es 2 Möglichkeiten: Entweder der 
        '  Node ist "richtig" (Übereinstimmung mit dem Dateisystem), und damit 
        '  nicht erneut hinzuzufügen. Oder er ist "falsch", und muss entfernt 
        '  werden.
        For INode As Integer = Nodes.Count - 1 To 0 Step -1
            Dim IName As Integer = Names.IndexOf(Nodes(INode).Text)
            If IName >= 0 Then
                ' "Richtigen" Node vom Hinzufügen ausnehmen
                Names.RemoveAt(IName)
            Else
                ' Ungültigen Node entfernen
                Nodes.RemoveAt(INode)
            End If
        Next

        ' Aus allen anderen Names Treenodes erzeugen und hinzufügen
        For Each sDir As String In Names
            ' Der neue Node ist als unsynchronisiert markiert, indem .Tag 
            '  irgendetwas zugewiesen ist
            Nodes.Add(New TreeNode(sDir) With {.Tag = 1})
        Next
    End Sub

    Public Shared Sub Fill(ByVal Node As TreeNode)
        ' Befüllung ohne Threading
        Synchronize(Node.Nodes, GetSubFolders(Node.FullPath))
        Node.Tag = Nothing ' als synchronisiert markieren
    End Sub

    Public Shared Sub FillWithDrives(ByVal Nodes As TreeNodeCollection)
        ' Befüllung ohne Threading, mit Laufwerken statt Verzeichnissen
        Dim SubFolders = From Info In DriveInfo.GetDrives _
                          Where Info.IsReady _
                          Select Info.Name.TrimEnd(Path.DirectorySeparatorChar)
        Synchronize(Nodes, SubFolders.ToList)
    End Sub

End Class
