﻿
''' <summary>
'''  Kapselt einen Queue Of Action-Delegaten, welcher im Nebenthread 
'''   abgearbeitet werden
''' </summary>
''' <remarks> 
'''  Während der Bearbeitung können weitere "Jobs" hinzugefügt werden 
''' </remarks>
Public Class JobWorker
    Public Class EventArg : Inherits EventArgs
        ''' <summary> 
        '''  Liste der Objekte, deren Objekt-Funktion abgearbeitet wurde 
        ''' </summary>
        Public ReadOnly DoneJobs As New List(Of Object)
    End Class

    Public Event AllDone As EventHandler(Of EventArg)
    Private _EA As New EventArg

    Private ReadOnly _Monitor As New DisposeLevel(AddressOf EnterMonitor, _
                                                  AddressOf ExitMonitor)
    Private ReadOnly _ToDo As New Queue(Of Action)
    Private _DoWork As Action = AddressOf DoWork
    Private _DoWorkCallback As AsyncCallback = AddressOf _DoWork.EndInvoke
    Private _WorkCompleted As Action = AddressOf WorkCompleted
    Private _IsBusy As Boolean

    Private Sub EnterMonitor()
        System.Threading.Monitor.Enter(_Monitor)
    End Sub
    Private Sub ExitMonitor()
        System.Threading.Monitor.Exit(_Monitor)
    End Sub

    ''' <summary>
    '''  Using JobWorker.Monitor reserviert die Queue des JobWorkers für den 
    '''  aktuellen Thread. So können mehrere Jobs zugefügt werden, ohne dass 
    '''  der JobWorker "dazwischenfunkt", indem er die Abarbeitung beginnt. 
    '''  Bei verschachtelten Blöcken wird die Queue erst beim Verlassen des 
    '''  äußeren Using-Blocks freigegeben.
    ''' </summary>
    Public ReadOnly Property Monitor() As IDisposable
        Get
            Return _Monitor.Enter
        End Get
    End Property

    ''' <summary>
    ''' Fügt einen Action - Delegaten hinzu, der abzuarbeiten ist. Startet 
    '''  gegebenenfalls den NebenThread.
    ''' </summary>
    Public Sub Add(ByVal Method As Action)
        ' SyncLock/Monitor.Enter sichert Zugriffe auf die Queue dagegen ab, dass
        '  gleichzeitig aus dem anderen Thread zugegriffen wird.
        Using _Monitor.Enter
            _ToDo.Enqueue(Method)
        End Using
        If Not _IsBusy Then
            _IsBusy = True
            _DoWork.BeginInvoke(_DoWorkCallback, Nothing)
        End If
    End Sub

    Private Sub DoWork()
        'Arbeiten im NebenThread
        While True
            Dim Method As Action = Nothing
            ' Falls der HauptThread grade Jobs hinzufügt, so lange warten
            SyncLock _Monitor
                If _ToDo.Count = 0 Then Exit While
                Method = _ToDo.Dequeue
            End SyncLock
            Method()
            ' Die Targets der Delegaten sind diejenigen NodeFillJobs, deren 
            '  .FillChildNames()-Funktion der Delegat aufruft. Die NodeFillJobs 
            '  sollen später im Hauptthread zum Abschluss gebracht werden. Sie 
            '  werden schonmal ins EventArg eingespeichert, welches später 
            '  verwendet wird.
            _EA.DoneJobs.Add(Method.Target)
        End While
        Application.OpenForms(0).BeginInvoke(_WorkCompleted)
    End Sub

    Private Sub WorkCompleted()
        If _ToDo.Count > 0 Then
            ' Mögliche Race-Condition: 
            '  während Application.OpenForms(0).BeginInvoke(_WorkCompleted) 
            '  "unterwegs" zum GuiThread ist, können weitere Jobs aufgelaufen 
            '  sein.
            _DoWork.BeginInvoke(_DoWorkCallback, Nothing)
            Return
        End If

        ' Eventuell können innerhalb der Behandlung von _AllDone neue Jobs 
        '  hinzugefügt werden, daher ein neues _EA anlegen.
        Dim OldEA As EventArg = _EA
        _EA = New EventArg
        _IsBusy = False
        RaiseEvent AllDone(Me, OldEA)
    End Sub

End Class
